@@ -1,6 +1,7 @@
 use std::{thread, time::Duration};
 
 use pyo3::prelude::*;
+use pyo3_asyncio::TaskLocals;
 
 pub(super) const TEST_MOD: &'static str = r#"
 import asyncio 
@@ -17,8 +18,8 @@ pub(super) async fn test_into_future(event_loop: PyObject) -> PyResult<()> {
         let test_mod =
             PyModule::from_code(py, TEST_MOD, "test_rust_coroutine/test_mod.py", "test_mod")?;
 
-        pyo3_asyncio::into_future_with_loop(
-            event_loop.as_ref(py),
+        pyo3_asyncio::into_future_with_locals(
+            &TaskLocals::new(event_loop.as_ref(py)),
             test_mod.call_method1("py_sleep", (1.into_py(py),))?,
         )
     })?;
@@ -28,20 +29,6 @@ pub(super) async fn test_into_future(event_loop: PyObject) -> PyResult<()> {
     Ok(())
 }
 
-#[allow(deprecated)]
-pub(super) async fn test_into_future_0_13() -> PyResult<()> {
-    let fut = Python::with_gil(|py| {
-        let test_mod =
-            PyModule::from_code(py, TEST_MOD, "test_rust_coroutine/test_mod.py", "test_mod")?;
-
-        pyo3_asyncio::into_future(test_mod.call_method1("py_sleep", (1.into_py(py),))?)
-    })?;
-
-    fut.await?;
-
-    Ok(())
-}
-
 pub(super) fn test_blocking_sleep() -> PyResult<()> {
     thread::sleep(Duration::from_secs(1));
     Ok(())
@@ -61,7 +48,7 @@ pub(super) async fn test_other_awaitables(event_loop: PyObject) -> PyResult<()>
             ),
         )?;
 
-        pyo3_asyncio::into_future_with_loop(event_loop.as_ref(py), task)
+        pyo3_asyncio::into_future_with_locals(&TaskLocals::new(event_loop.as_ref(py)), task)
     })?;
 
     fut.await?;
