@@ -1,7 +1,7 @@
-use pyo3::{prelude::*, types::PyType};
-
 #[cfg(not(target_os = "windows"))]
-fn main() -> PyResult<()> {
+fn main() -> pyo3::PyResult<()> {
+    use pyo3::{prelude::*, types::PyType};
+
     pyo3::prepare_freethreaded_python();
 
     let mut builder = tokio::runtime::Builder::new_current_thread();
