@@ -4,7 +4,10 @@ use async_std::task;
 use futures::prelude::*;
 use pyo3::prelude::*;
 
-use crate::generic::{self, JoinError, Runtime, SpawnLocalExt};
+use crate::{
+    generic::{self, ContextExt, JoinError, LocalContextExt, Runtime, SpawnLocalExt},
+    TaskLocals,
+};
 
 /// <span class="module-item stab portability" style="display: inline; border-radius: 3px; padding: 2px; font-size: 80%; line-height: 1.2;"><code>attributes</code></span>
 /// re-exports for macros
@@ -33,7 +36,7 @@ impl JoinError for AsyncStdJoinErr {
 }
 
 async_std::task_local! {
-    static EVENT_LOOP: RefCell<Option<PyObject>> = RefCell::new(None);
+    static TASK_LOCALS: RefCell<Option<TaskLocals>> = RefCell::new(None);
 }
 
 struct AsyncStdRuntime;
@@ -42,28 +45,6 @@ impl Runtime for AsyncStdRuntime {
     type JoinError = AsyncStdJoinErr;
     type JoinHandle = task::JoinHandle<Result<(), AsyncStdJoinErr>>;
 
-    fn scope<F, R>(event_loop: PyObject, fut: F) -> Pin<Box<dyn Future<Output = R> + Send>>
-    where
-        F: Future<Output = R> + Send + 'static,
-    {
-        let old = EVENT_LOOP.with(|c| c.replace(Some(event_loop)));
-        Box::pin(async move {
-            let result = fut.await;
-            EVENT_LOOP.with(|c| c.replace(old));
-            result
-        })
-    }
-    fn get_task_event_loop(py: Python) -> Option<&PyAny> {
-        match EVENT_LOOP.try_with(|c| {
-            c.borrow()
-                .as_ref()
-                .map(|event_loop| event_loop.clone().into_ref(py))
-        }) {
-            Ok(event_loop) => event_loop,
-            Err(_) => None,
-        }
-    }
-
     fn spawn<F>(fut: F) -> Self::JoinHandle
     where
         F: Future<Output = ()> + Send + 'static,
@@ -77,19 +58,28 @@ impl Runtime for AsyncStdRuntime {
     }
 }
 
-impl SpawnLocalExt for AsyncStdRuntime {
-    fn scope_local<F, R>(event_loop: PyObject, fut: F) -> Pin<Box<dyn Future<Output = R>>>
+impl ContextExt for AsyncStdRuntime {
+    fn scope<F, R>(locals: TaskLocals, fut: F) -> Pin<Box<dyn Future<Output = R> + Send>>
     where
-        F: Future<Output = R> + 'static,
+        F: Future<Output = R> + Send + 'static,
     {
-        let old = EVENT_LOOP.with(|c| c.replace(Some(event_loop)));
+        let old = TASK_LOCALS.with(|c| c.replace(Some(locals)));
         Box::pin(async move {
             let result = fut.await;
-            EVENT_LOOP.with(|c| c.replace(old));
+            TASK_LOCALS.with(|c| c.replace(old));
             result
         })
     }
 
+    fn get_task_locals() -> Option<TaskLocals> {
+        match TASK_LOCALS.try_with(|c| c.borrow().clone()) {
+            Ok(locals) => locals,
+            Err(_) => None,
+        }
+    }
+}
+
+impl SpawnLocalExt for AsyncStdRuntime {
     fn spawn_local<F>(fut: F) -> Self::JoinHandle
     where
         F: Future<Output = ()> + 'static,
@@ -101,20 +91,34 @@ impl SpawnLocalExt for AsyncStdRuntime {
     }
 }
 
+impl LocalContextExt for AsyncStdRuntime {
+    fn scope_local<F, R>(locals: TaskLocals, fut: F) -> Pin<Box<dyn Future<Output = R>>>
+    where
+        F: Future<Output = R> + 'static,
+    {
+        let old = TASK_LOCALS.with(|c| c.replace(Some(locals)));
+        Box::pin(async move {
+            let result = fut.await;
+            TASK_LOCALS.with(|c| c.replace(old));
+            result
+        })
+    }
+}
+
 /// Set the task local event loop for the given future
-pub async fn scope<F, R>(event_loop: PyObject, fut: F) -> R
+pub async fn scope<F, R>(locals: TaskLocals, fut: F) -> R
 where
     F: Future<Output = R> + Send + 'static,
 {
-    AsyncStdRuntime::scope(event_loop, fut).await
+    AsyncStdRuntime::scope(locals, fut).await
 }
 
 /// Set the task local event loop for the given !Send future
-pub async fn scope_local<F, R>(event_loop: PyObject, fut: F) -> R
+pub async fn scope_local<F, R>(locals: TaskLocals, fut: F) -> R
 where
     F: Future<Output = R> + 'static,
 {
-    AsyncStdRuntime::scope_local(event_loop, fut).await
+    AsyncStdRuntime::scope_local(locals, fut).await
 }
 
 /// Get the current event loop from either Python or Rust async task local context
@@ -126,12 +130,17 @@ pub fn get_current_loop(py: Python) -> PyResult<&PyAny> {
     generic::get_current_loop::<AsyncStdRuntime>(py)
 }
 
+/// Either copy the task locals from the current task OR get the current running loop and
+/// contextvars from Python.
+pub fn get_current_locals(py: Python) -> PyResult<TaskLocals> {
+    generic::get_current_locals::<AsyncStdRuntime>(py)
+}
+
 /// Run the event loop until the given Future completes
 ///
 /// The event loop runs until the given future is complete.
 ///
-/// After this function returns, the event loop can be resumed with either [`run_until_complete`] or
-/// [`run_forever`](`crate::run_forever`)
+/// After this function returns, the event loop can be resumed with [`run_until_complete`]
 ///
 /// # Arguments
 /// * `event_loop` - The Python event loop that should run the future
@@ -146,26 +155,21 @@ pub fn get_current_loop(py: Python) -> PyResult<&PyAny> {
 /// #
 /// # pyo3::prepare_freethreaded_python();
 /// #
-/// # Python::with_gil(|py| {
-/// # pyo3_asyncio::with_runtime(py, || {
+/// # Python::with_gil(|py| -> PyResult<()> {
 /// # let event_loop = py.import("asyncio")?.call_method0("new_event_loop")?;
 /// pyo3_asyncio::async_std::run_until_complete(event_loop, async move {
 ///     async_std::task::sleep(Duration::from_secs(1)).await;
 ///     Ok(())
 /// })?;
 /// # Ok(())
-/// # })
-/// # .map_err(|e| {
-/// #    e.print_and_set_sys_last_vars(py);  
-/// # })
-/// # .unwrap();
-/// # });
+/// # }).unwrap();
 /// ```
-pub fn run_until_complete<F>(event_loop: &PyAny, fut: F) -> PyResult<()>
+pub fn run_until_complete<F, T>(event_loop: &PyAny, fut: F) -> PyResult<T>
 where
-    F: Future<Output = PyResult<()>> + Send + 'static,
+    F: Future<Output = PyResult<T>> + Send + 'static,
+    T: Send + Sync + 'static,
 {
-    generic::run_until_complete::<AsyncStdRuntime, _>(event_loop, fut)
+    generic::run_until_complete::<AsyncStdRuntime, _, T>(event_loop, fut)
 }
 
 /// Run the event loop until the given Future completes
@@ -197,50 +201,17 @@ where
 ///     })
 /// }
 /// ```
-pub fn run<F>(py: Python, fut: F) -> PyResult<()>
+pub fn run<F, T>(py: Python, fut: F) -> PyResult<T>
 where
-    F: Future<Output = PyResult<()>> + Send + 'static,
+    F: Future<Output = PyResult<T>> + Send + 'static,
+    T: Send + Sync + 'static,
 {
-    generic::run::<AsyncStdRuntime, F>(py, fut)
+    generic::run::<AsyncStdRuntime, F, T>(py, fut)
 }
 
 /// Convert a Rust Future into a Python awaitable
 ///
-/// # Arguments
-/// * `py` - The current PyO3 GIL guard
-/// * `fut` - The Rust future to be converted
-///
-/// # Examples
-///
-/// ```
-/// use std::time::Duration;
-///
-/// use pyo3::prelude::*;
-///
-/// /// Awaitable sleep function
-/// #[pyfunction]
-/// fn sleep_for(py: Python, secs: &PyAny) -> PyResult<PyObject> {
-///     let secs = secs.extract()?;
-///
-///     pyo3_asyncio::async_std::into_coroutine(py, async move {
-///         async_std::task::sleep(Duration::from_secs(secs)).await;
-///         Python::with_gil(|py| Ok(py.None()))
-///     })
-/// }
-/// ```
-#[deprecated(
-    since = "0.14.0",
-    note = "Use the pyo3_asyncio::async_std::future_into_py instead\n    (see the [migration guide](https://github.com/awestlake87/pyo3-asyncio/#migrating-from-013-to-014) for more details)"
-)]
-#[allow(deprecated)]
-pub fn into_coroutine<F>(py: Python, fut: F) -> PyResult<PyObject>
-where
-    F: Future<Output = PyResult<PyObject>> + Send + 'static,
-{
-    generic::into_coroutine::<AsyncStdRuntime, _>(py, fut)
-}
-
-/// Convert a Rust Future into a Python awaitable
+/// __This function will be removed in `v0.16`__
 ///
 /// # Arguments
 /// * `event_loop` - The Python event loop that the awaitable should be attached to
@@ -266,6 +237,11 @@ where
 ///     )
 /// }
 /// ```
+#[deprecated(
+    since = "0.15.0",
+    note = "Use pyo3_asyncio::async_std::future_into_py_with_locals instead"
+)]
+#[allow(deprecated)]
 pub fn future_into_py_with_loop<F>(event_loop: &PyAny, fut: F) -> PyResult<&PyAny>
 where
     F: Future<Output = PyResult<PyObject>> + Send + 'static,
@@ -275,12 +251,62 @@ where
 
 /// Convert a Rust Future into a Python awaitable
 ///
-/// Unlike [`future_into_py_with_loop`], this function will stop the Rust future from running when
-/// the `asyncio.Future` is cancelled from Python.
+/// If the `asyncio.Future` returned by this conversion is cancelled via `asyncio.Future.cancel`,
+/// the Rust future will be cancelled as well (new behaviour in `v0.15`).
 ///
-/// __This function will be deprecated in favor of [`future_into_py_with_loop`] in `v0.15` because
-/// it will become the default behaviour. In `v0.15`, any calls to this function can be seamlessly
-/// replaced with [`future_into_py_with_loop`].__
+/// Python `contextvars` are preserved when calling async Python functions within the Rust future
+/// via [`into_future`] (new behaviour in `v0.15`).
+///
+/// > Although `contextvars` are preserved for async Python functions, synchronous functions will
+/// unfortunately fail to resolve them when called within the Rust future. This is because the
+/// function is being called from a Rust thread, not inside an actual Python coroutine context.
+/// >
+/// > As a workaround, you can get the `contextvars` from the current task locals using
+/// [`get_current_locals`] and [`TaskLocals::context`](`crate::TaskLocals::context`), then wrap your
+/// synchronous function in a call to `contextvars.Context.run`. This will set the context, call the
+/// synchronous function, and restore the previous context when it returns or raises an exception.
+///
+/// # Arguments
+/// * `py` - PyO3 GIL guard
+/// * `locals` - The task locals for the given future
+/// * `fut` - The Rust future to be converted
+///
+/// # Examples
+///
+/// ```
+/// use std::time::Duration;
+///
+/// use pyo3::prelude::*;
+///
+/// /// Awaitable sleep function
+/// #[pyfunction]
+/// fn sleep_for<'p>(py: Python<'p>, secs: &'p PyAny) -> PyResult<&'p PyAny> {
+///     let secs = secs.extract()?;
+///     pyo3_asyncio::async_std::future_into_py_with_locals(
+///         py,
+///         pyo3_asyncio::async_std::get_current_locals(py)?,
+///         async move {
+///             async_std::task::sleep(Duration::from_secs(secs)).await;
+///             Python::with_gil(|py| Ok(py.None()))
+///         }
+///     )
+/// }
+/// ```
+pub fn future_into_py_with_locals<F, T>(py: Python, locals: TaskLocals, fut: F) -> PyResult<&PyAny>
+where
+    F: Future<Output = PyResult<T>> + Send + 'static,
+    T: IntoPy<PyObject>,
+{
+    generic::future_into_py_with_locals::<AsyncStdRuntime, F, T>(py, locals, fut)
+}
+
+/// Convert a Rust Future into a Python awaitable
+///
+/// __This function was deprecated in favor of [`future_into_py_with_locals`] in `v0.15` because
+/// it became the default behaviour. In `v0.15`, any calls to this function should be replaced with
+/// [`future_into_py_with_locals`]__
+///
+/// __This function will be removed in `v0.16`__
 ///
 /// # Arguments
 /// * `event_loop` - The Python event loop that the awaitable should be attached to
@@ -301,11 +327,16 @@ where
 ///         pyo3_asyncio::async_std::get_current_loop(py)?,
 ///         async move {
 ///             async_std::task::sleep(Duration::from_secs(secs)).await;
-///             Python::with_gil(|py| Ok(py.None()))
+///             Ok(Python::with_gil(|py| py.None()))
 ///         }
 ///     )
 /// }
 /// ```
+#[deprecated(
+    since = "0.15.0",
+    note = "Use pyo3_asyncio::async_std::future_into_py_with_locals instead"
+)]
+#[allow(deprecated)]
 pub fn cancellable_future_into_py_with_loop<F>(event_loop: &PyAny, fut: F) -> PyResult<&PyAny>
 where
     F: Future<Output = PyResult<PyObject>> + Send + 'static,
@@ -315,6 +346,21 @@ where
 
 /// Convert a Rust Future into a Python awaitable
 ///
+/// If the `asyncio.Future` returned by this conversion is cancelled via `asyncio.Future.cancel`,
+/// the Rust future will be cancelled as well (new behaviour in `v0.15`).
+///
+/// Python `contextvars` are preserved when calling async Python functions within the Rust future
+/// via [`into_future`] (new behaviour in `v0.15`).
+///
+/// > Although `contextvars` are preserved for async Python functions, synchronous functions will
+/// unfortunately fail to resolve them when called within the Rust future. This is because the
+/// function is being called from a Rust thread, not inside an actual Python coroutine context.
+/// >
+/// > As a workaround, you can get the `contextvars` from the current task locals using
+/// [`get_current_locals`] and [`TaskLocals::context`](`crate::TaskLocals::context`), then wrap your
+/// synchronous function in a call to `contextvars.Context.run`. This will set the context, call the
+/// synchronous function, and restore the previous context when it returns or raises an exception.
+///
 /// # Arguments
 /// * `py` - The current PyO3 GIL guard
 /// * `fut` - The Rust future to be converted
@@ -332,26 +378,26 @@ where
 ///     let secs = secs.extract()?;
 ///     pyo3_asyncio::async_std::future_into_py(py, async move {
 ///         async_std::task::sleep(Duration::from_secs(secs)).await;
-///         Python::with_gil(|py| Ok(py.None()))
+///         Ok(())
 ///     })
 /// }
 /// ```
-pub fn future_into_py<F>(py: Python, fut: F) -> PyResult<&PyAny>
+pub fn future_into_py<F, T>(py: Python, fut: F) -> PyResult<&PyAny>
 where
-    F: Future<Output = PyResult<PyObject>> + Send + 'static,
+    F: Future<Output = PyResult<T>> + Send + 'static,
+    T: IntoPy<PyObject>,
 {
-    generic::future_into_py::<AsyncStdRuntime, _>(py, fut)
+    generic::future_into_py::<AsyncStdRuntime, _, T>(py, fut)
 }
 
 /// Convert a Rust Future into a Python awaitable
 ///
-/// Unlike [`future_into_py`], this function will stop the Rust future from running when
-/// the `asyncio.Future` is cancelled from Python.
-///
-/// __This function will be deprecated in favor of [`future_into_py`] in `v0.15` because
-/// it will become the default behaviour. In `v0.15`, any calls to this function can be seamlessly
+/// __This function was deprecated in favor of [`future_into_py`] in `v0.15` because
+/// it became the default behaviour. In `v0.15`, any calls to this function can be seamlessly
 /// replaced with [`future_into_py`].__
 ///
+/// __This function will be removed in `v0.16`__
+///
 /// # Arguments
 /// * `py` - The current PyO3 GIL guard
 /// * `fut` - The Rust future to be converted
@@ -373,6 +419,11 @@ where
 ///     })
 /// }
 /// ```
+#[deprecated(
+    since = "0.15.0",
+    note = "Use pyo3_asyncio::async_std::future_into_py instead"
+)]
+#[allow(deprecated)]
 pub fn cancellable_future_into_py<F>(py: Python, fut: F) -> PyResult<&PyAny>
 where
     F: Future<Output = PyResult<PyObject>> + Send + 'static,
@@ -382,6 +433,8 @@ where
 
 /// Convert a `!Send` Rust Future into a Python awaitable
 ///
+/// __This function will be removed in `v0.16`__
+///
 /// # Arguments
 /// * `event_loop` - The Python event loop that the awaitable should be attached to
 /// * `fut` - The Rust future to be converted
@@ -402,6 +455,76 @@ where
 ///         pyo3_asyncio::async_std::get_current_loop(py)?,
 ///         async move {
 ///             async_std::task::sleep(Duration::from_secs(*secs)).await;
+///             Ok(Python::with_gil(|py| py.None()))
+///         }
+///     )?.into())
+/// }
+///
+/// # #[cfg(all(feature = "async-std-runtime", feature = "attributes"))]
+/// #[pyo3_asyncio::async_std::main]
+/// async fn main() -> PyResult<()> {
+///     Python::with_gil(|py| {
+///         let py_future = sleep_for(py, 1)?;
+///         pyo3_asyncio::async_std::into_future(py_future)
+///     })?
+///     .await?;
+///
+///     Ok(())
+/// }
+/// # #[cfg(not(all(feature = "async-std-runtime", feature = "attributes")))]
+/// # fn main() {}
+/// ```
+#[deprecated(
+    since = "0.15.0",
+    note = "Use pyo3_asyncio::async_std::local_future_into_py_with_locals instead"
+)]
+#[allow(deprecated)]
+pub fn local_future_into_py_with_loop<F>(event_loop: &PyAny, fut: F) -> PyResult<&PyAny>
+where
+    F: Future<Output = PyResult<PyObject>> + 'static,
+{
+    generic::local_future_into_py_with_loop::<AsyncStdRuntime, _>(event_loop, fut)
+}
+
+/// Convert a `!Send` Rust Future into a Python awaitable
+///
+/// If the `asyncio.Future` returned by this conversion is cancelled via `asyncio.Future.cancel`,
+/// the Rust future will be cancelled as well (new behaviour in `v0.15`).
+///
+/// Python `contextvars` are preserved when calling async Python functions within the Rust future
+/// via [`into_future`] (new behaviour in `v0.15`).
+///
+/// > Although `contextvars` are preserved for async Python functions, synchronous functions will
+/// unfortunately fail to resolve them when called within the Rust future. This is because the
+/// function is being called from a Rust thread, not inside an actual Python coroutine context.
+/// >
+/// > As a workaround, you can get the `contextvars` from the current task locals using
+/// [`get_current_locals`] and [`TaskLocals::context`](`crate::TaskLocals::context`), then wrap your
+/// synchronous function in a call to `contextvars.Context.run`. This will set the context, call the
+/// synchronous function, and restore the previous context when it returns or raises an exception.
+///
+/// # Arguments
+/// * `py` - PyO3 GIL guard
+/// * `locals` - The task locals for the given future
+/// * `fut` - The Rust future to be converted
+///
+/// # Examples
+///
+/// ```
+/// use std::{rc::Rc, time::Duration};
+///
+/// use pyo3::prelude::*;
+///
+/// /// Awaitable non-send sleep function
+/// #[pyfunction]
+/// fn sleep_for(py: Python, secs: u64) -> PyResult<&PyAny> {
+///     // Rc is non-send so it cannot be passed into pyo3_asyncio::async_std::future_into_py
+///     let secs = Rc::new(secs);
+///     Ok(pyo3_asyncio::async_std::local_future_into_py_with_locals(
+///         py,
+///         pyo3_asyncio::async_std::get_current_locals(py)?,
+///         async move {
+///             async_std::task::sleep(Duration::from_secs(*secs)).await;
 ///             Python::with_gil(|py| Ok(py.None()))
 ///         }
 ///     )?.into())
@@ -421,21 +544,25 @@ where
 /// # #[cfg(not(all(feature = "async-std-runtime", feature = "attributes")))]
 /// # fn main() {}
 /// ```
-pub fn local_future_into_py_with_loop<F>(event_loop: &PyAny, fut: F) -> PyResult<&PyAny>
+pub fn local_future_into_py_with_locals<F, T>(
+    py: Python,
+    locals: TaskLocals,
+    fut: F,
+) -> PyResult<&PyAny>
 where
-    F: Future<Output = PyResult<PyObject>> + 'static,
+    F: Future<Output = PyResult<T>> + 'static,
+    T: IntoPy<PyObject>,
 {
-    generic::local_future_into_py_with_loop::<AsyncStdRuntime, _>(event_loop, fut)
+    generic::local_future_into_py_with_locals::<AsyncStdRuntime, _, T>(py, locals, fut)
 }
 
 /// Convert a `!Send` Rust Future into a Python awaitable
 ///
-/// Unlike [`local_future_into_py_with_loop`], this function will stop the Rust future from running when
-/// the `asyncio.Future` is cancelled from Python.
+/// __This function was deprecated in favor of [`local_future_into_py_with_locals`] in `v0.15` because
+/// it became the default behaviour. In `v0.15`, any calls to this function should be
+/// replaced with [`local_future_into_py_with_locals`].__
 ///
-/// __This function will be deprecated in favor of [`local_future_into_py_with_loop`] in `v0.15` because
-/// it will become the default behaviour. In `v0.15`, any calls to this function can be seamlessly
-/// replaced with [`local_future_into_py_with_loop`].__
+/// __This function will be removed in `v0.16`__
 ///
 /// # Arguments
 /// * `event_loop` - The Python event loop that the awaitable should be attached to
@@ -476,6 +603,11 @@ where
 /// # #[cfg(not(all(feature = "async-std-runtime", feature = "attributes")))]
 /// # fn main() {}
 /// ```
+#[deprecated(
+    since = "0.15.0",
+    note = "Use pyo3_asyncio::async_std::local_future_into_py_with_locals instead"
+)]
+#[allow(deprecated)]
 pub fn local_cancellable_future_into_py_with_loop<F>(event_loop: &PyAny, fut: F) -> PyResult<&PyAny>
 where
     F: Future<Output = PyResult<PyObject>> + 'static,
@@ -485,6 +617,21 @@ where
 
 /// Convert a `!Send` Rust Future into a Python awaitable
 ///
+/// If the `asyncio.Future` returned by this conversion is cancelled via `asyncio.Future.cancel`,
+/// the Rust future will be cancelled as well (new behaviour in `v0.15`).
+///
+/// Python `contextvars` are preserved when calling async Python functions within the Rust future
+/// via [`into_future`] (new behaviour in `v0.15`).
+///
+/// > Although `contextvars` are preserved for async Python functions, synchronous functions will
+/// unfortunately fail to resolve them when called within the Rust future. This is because the
+/// function is being called from a Rust thread, not inside an actual Python coroutine context.
+/// >
+/// > As a workaround, you can get the `contextvars` from the current task locals using
+/// [`get_current_locals`] and [`TaskLocals::context`](`crate::TaskLocals::context`), then wrap your
+/// synchronous function in a call to `contextvars.Context.run`. This will set the context, call the
+/// synchronous function, and restore the previous context when it returns or raises an exception.
+///
 /// # Arguments
 /// * `py` - The current PyO3 GIL guard
 /// * `fut` - The Rust future to be converted
@@ -503,7 +650,7 @@ where
 ///     let secs = Rc::new(secs);
 ///     pyo3_asyncio::async_std::local_future_into_py(py, async move {
 ///         async_std::task::sleep(Duration::from_secs(*secs)).await;
-///         Python::with_gil(|py| Ok(py.None()))
+///         Ok(())
 ///     })
 /// }
 ///
@@ -521,22 +668,22 @@ where
 /// # #[cfg(not(all(feature = "async-std-runtime", feature = "attributes")))]
 /// # fn main() {}
 /// ```
-pub fn local_future_into_py<F>(py: Python, fut: F) -> PyResult<&PyAny>
+pub fn local_future_into_py<F, T>(py: Python, fut: F) -> PyResult<&PyAny>
 where
-    F: Future<Output = PyResult<PyObject>> + 'static,
+    F: Future<Output = PyResult<T>> + 'static,
+    T: IntoPy<PyObject>,
 {
-    generic::local_future_into_py::<AsyncStdRuntime, _>(py, fut)
+    generic::local_future_into_py::<AsyncStdRuntime, _, T>(py, fut)
 }
 
 /// Convert a `!Send` Rust Future into a Python awaitable
 ///
-/// Unlike [`local_future_into_py`], this function will stop the Rust future from running when
-/// the `asyncio.Future` is cancelled from Python.
-///
-/// __This function will be deprecated in favor of [`local_future_into_py`] in `v0.15` because
-/// it will become the default behaviour. In `v0.15`, any calls to this function can be seamlessly
+/// __This function was deprecated in favor of [`local_future_into_py`] in `v0.15` because
+/// it became the default behaviour. In `v0.15`, any calls to this function can be seamlessly
 /// replaced with [`local_future_into_py`].__
 ///
+/// __This function will be removed in `v0.16`__
+///
 /// # Arguments
 /// * `py` - The current PyO3 GIL guard
 /// * `fut` - The Rust future to be converted
@@ -573,6 +720,11 @@ where
 /// # #[cfg(not(all(feature = "async-std-runtime", feature = "attributes")))]
 /// # fn main() {}
 /// ```
+#[deprecated(
+    since = "0.15.0",
+    note = "Use pyo3_asyncio::async_std::local_future_into_py instead"
+)]
+#[allow(deprecated)]
 pub fn local_cancellable_future_into_py<F>(py: Python, fut: F) -> PyResult<&PyAny>
 where
     F: Future<Output = PyResult<PyObject>> + 'static,
