@@ -18,37 +18,69 @@
 //! be run on the same loop.
 //!
 //! It's not immediately clear that this would provide worthwhile performance wins either, so in the
-//! interest of keeping things simple, this crate creates and manages the Python event loop and
-//! handles the communication between separate Rust event loops.
+//! interest of getting something simple out there to facilitate these conversions, this crate
+//! handles the communication between _separate_ Python and Rust event loops.
 //!
-//! ## Python's Event Loop
+//! ## Python's Event Loop and the Main Thread
 //!
 //! Python is very picky about the threads used by the `asyncio` executor. In particular, it needs
 //! to have control over the main thread in order to handle signals like CTRL-C correctly. This
 //! means that Cargo's default test harness will no longer work since it doesn't provide a method of
 //! overriding the main function to add our event loop initialization and finalization.
 //!
-//! ## Event Loop References
+//! ## Event Loop References and ContextVars
 //!
-//! One problem that arises when interacting with Python's asyncio library is that the functions we use to get a reference to the Python event loop can only be called in certain contexts. Since PyO3 Asyncio needs to interact with Python's event loop during conversions, the context of these conversions can matter a lot.
+//! One problem that arises when interacting with Python's asyncio library is that the functions we
+//! use to get a reference to the Python event loop can only be called in certain contexts. Since
+//! PyO3 Asyncio needs to interact with Python's event loop during conversions, the context of these
+//! conversions can matter a lot.
 //!
-//! > The core conversions we've mentioned so far in this guide should insulate you from these concerns in most cases, but in the event that they don't, this section should provide you with the information you need to solve these problems.
+//! Likewise, Python's `contextvars` library can require some special treatment. Python functions
+//! and coroutines can rely on the context of outer coroutines to function correctly, so this
+//! library needs to be able to preserve `contextvars` during conversions.
+//!
+//! > The core conversions we've mentioned so far in the README should insulate you from these
+//! concerns in most cases. For the edge cases where they don't, this section should provide you
+//! with the information you need to solve these problems.
 //!
 //! ### The Main Dilemma
 //!
-//! Python programs can have many independent event loop instances throughout the lifetime of the application (`asyncio.run` for example creates its own event loop each time it's called for instance), and they can even run concurrent with other event loops. For this reason, the most correct method of obtaining a reference to the Python event loop is via `asyncio.get_running_loop`.
+//! Python programs can have many independent event loop instances throughout the lifetime of the
+//! application (`asyncio.run` for example creates its own event loop each time it's called for
+//! instance), and they can even run concurrent with other event loops. For this reason, the most
+//! correct method of obtaining a reference to the Python event loop is via
+//! `asyncio.get_running_loop`.
 //!
-//! `asyncio.get_running_loop` returns the event loop associated with the current OS thread. It can be used inside Python coroutines to spawn concurrent tasks, interact with timers, or in our case signal between Rust and Python. This is all well and good when we are operating on a Python thread, but since Rust threads are not associated with a Python event loop, `asyncio.get_running_loop` will fail when called on a Rust runtime.
+//! `asyncio.get_running_loop` returns the event loop associated with the current OS thread. It can
+//! be used inside Python coroutines to spawn concurrent tasks, interact with timers, or in our case
+//! signal between Rust and Python. This is all well and good when we are operating on a Python
+//! thread, but since Rust threads are not associated with a Python event loop,
+//! `asyncio.get_running_loop` will fail when called on a Rust runtime.
+//!
+//! `contextvars` operates in a similar way, though the current context is not always associated
+//! with the current OS thread. Different contexts can be associated with different coroutines even
+//! if they run on the same OS thread.
 //!
 //! ### The Solution
 //!
-//! A really straightforward way of dealing with this problem is to pass a reference to the associated Python event loop for every conversion. That's why in `v0.14`, we introduced a new set of conversion functions that do just that:
+//! A really straightforward way of dealing with this problem is to pass references to the
+//! associated Python event loop and context for every conversion. That's why we have a structure
+//! called `TaskLocals` and a set of conversions that accept it.
 //!
-//! - `pyo3_asyncio::into_future_with_loop` - Convert a Python awaitable into a Rust future with the given asyncio event loop.
-//! - `pyo3_asyncio::<runtime>::future_into_py_with_loop` - Convert a Rust future into a Python awaitable with the given asyncio event loop.
-//! - `pyo3_asyncio::<runtime>::local_future_into_py_with_loop` - Convert a `!Send` Rust future into a Python awaitable with the given asyncio event loop.
+//! `TaskLocals` stores the current event loop, and allows the user to copy the current Python
+//! context if necessary. The following conversions will use these references to perform the
+//! necessary conversions and restore Python context when needed:
 //!
-//! One clear disadvantage to this approach (aside from the verbose naming) is that the Rust application has to explicitly track its references to the Python event loop. In native libraries, we can't make any assumptions about the underlying event loop, so the only reliable way to make sure our conversions work properly is to store a reference to the current event loop at the callsite to use later on.
+//! - `pyo3_asyncio::into_future_with_locals` - Convert a Python awaitable into a Rust future.
+//! - `pyo3_asyncio::<runtime>::future_into_py_with_locals` - Convert a Rust future into a Python
+//! awaitable.
+//! - `pyo3_asyncio::<runtime>::local_future_into_py_with_locals` - Convert a `!Send` Rust future
+//! into a Python awaitable.
+//!
+//! One clear disadvantage to this approach is that the Rust application has to explicitly track
+//! these references. In native libraries, we can't make any assumptions about the underlying event
+//! loop, so the only reliable way to make sure our conversions work properly is to store these
+//! references at the callsite to use later on.
 //!
 //! ```rust
 //! use pyo3::{wrap_pyfunction, prelude::*};
@@ -56,24 +88,24 @@
 //! # #[cfg(feature = "tokio-runtime")]
 //! #[pyfunction]
 //! fn sleep(py: Python) -> PyResult<&PyAny> {
-//!     let current_loop = pyo3_asyncio::get_running_loop(py)?;
-//!     let loop_ref = PyObject::from(current_loop);
+//!     // Construct the task locals structure with the current running loop and context
+//!     let locals = pyo3_asyncio::TaskLocals::with_running_loop(py)?.copy_context(py)?;
 //!
 //!     // Convert the async move { } block to a Python awaitable
-//!     pyo3_asyncio::tokio::future_into_py_with_loop(current_loop, async move {
+//!     pyo3_asyncio::tokio::future_into_py_with_locals(py, locals.clone(), async move {
 //!         let py_sleep = Python::with_gil(|py| {
 //!             // Sometimes we need to call other async Python functions within
 //!             // this future. In order for this to work, we need to track the
 //!             // event loop from earlier.
-//!             pyo3_asyncio::into_future_with_loop(
-//!                 loop_ref.as_ref(py),
+//!             pyo3_asyncio::into_future_with_locals(
+//!                 &locals,
 //!                 py.import("asyncio")?.call_method1("sleep", (1,))?
 //!             )
 //!         })?;
 //!
 //!         py_sleep.await?;
 //!
-//!         Ok(Python::with_gil(|py| py.None()))
+//!         Ok(())
 //!     })
 //! }
 //!
@@ -85,18 +117,35 @@
 //! }
 //! ```
 //!
-//! > A naive solution to this tracking problem would be to cache a global reference to the asyncio event loop that all PyO3 Asyncio conversions can use. In fact this is what we did in PyO3 Asyncio `v0.13`. This works well for applications, but it soon became clear that this is not so ideal for libraries. Libraries usually have no direct control over how the event loop is managed, they're just expected to work with any event loop at any point in the application. This problem is compounded further when multiple event loops are used in the application since the global reference will only point to one.
+//! > A naive solution to this tracking problem would be to cache a global reference to the asyncio
+//! event loop that all PyO3 Asyncio conversions can use. In fact this is what we did in PyO3
+//! Asyncio `v0.13`. This works well for applications, but it soon became clear that this is not
+//! so ideal for libraries. Libraries usually have no direct control over how the event loop is
+//! managed, they're just expected to work with any event loop at any point in the application.
+//! This problem is compounded further when multiple event loops are used in the application since
+//! the global reference will only point to one.
 //!
-//! Another disadvantage to this explicit approach that is less obvious is that we can no longer call our `#[pyfunction] fn sleep` on a Rust runtime since `asyncio.get_running_loop` only works on Python threads! It's clear that we need a slightly more flexible approach.
+//! Another disadvantage to this explicit approach that is less obvious is that we can no longer
+//! call our `#[pyfunction] fn sleep` on a Rust runtime since `asyncio.get_running_loop` only works
+//! on Python threads! It's clear that we need a slightly more flexible approach.
 //!
-//! In order to detect the Python event loop at the callsite, we need something like `asyncio.get_running_loop` that works for _both Python and Rust_. In Python, `asyncio.get_running_loop` uses thread-local data to retrieve the event loop associated with the current thread. What we need in Rust is something that can retrieve the Python event loop associated with the current _task_.
+//! In order to detect the Python event loop at the callsite, we need something like
+//! `asyncio.get_running_loop` and `contextvars.copy_context` that works for _both Python and Rust_.
+//! In Python, `asyncio.get_running_loop` uses thread-local data to retrieve the event loop
+//! associated with the current thread. What we need in Rust is something that can retrieve the
+//! Python event loop and contextvars associated with the current Rust _task_.
 //!
-//! Enter `pyo3_asyncio::<runtime>::get_current_loop`. This function first checks task-local data for a Python event loop, then falls back on `asyncio.get_running_loop` if no task-local event loop is found. This way both bases are covered.
+//! Enter `pyo3_asyncio::<runtime>::get_current_locals`. This function first checks task-local data
+//! for the `TaskLocals`, then falls back on `asyncio.get_running_loop` and
+//! `contextvars.copy_context` if no task locals are found. This way both bases are
+//! covered.
 //!
-//! Now, all we need is a way to store the event loop in task-local data. Since this is a runtime-specific feature, you can find the following functions in each runtime module:
+//! Now, all we need is a way to store the `TaskLocals` for the Rust future. Since this is a
+//! runtime-specific feature, you can find the following functions in each runtime module:
 //!
-//! - `pyo3_asyncio::<runtime>::scope` - Store the event loop in task-local data when executing the given Future.
-//! - `pyo3_asyncio::<runtime>::scope_local` - Store the event loop in task-local data when executing the given `!Send` Future.
+//! - `pyo3_asyncio::<runtime>::scope` - Store the task-local data when executing the given Future.
+//! - `pyo3_asyncio::<runtime>::scope_local` - Store the task-local data when executing the given
+//! `!Send` Future.
 //!
 //! With these new functions, we can make our previous example more correct:
 //!
@@ -107,17 +156,18 @@
 //! #[pyfunction]
 //! fn sleep(py: Python) -> PyResult<&PyAny> {
 //!     // get the current event loop through task-local data
-//!     // OR `asyncio.get_running_loop`
-//!     let current_loop = pyo3_asyncio::tokio::get_current_loop(py)?;
+//!     // OR `asyncio.get_running_loop` and `contextvars.copy_context`
+//!     let locals = pyo3_asyncio::tokio::get_current_locals(py)?;
 //!
-//!     pyo3_asyncio::tokio::future_into_py_with_loop(
-//!         current_loop,
-//!         // Store the current loop in task-local data
-//!         pyo3_asyncio::tokio::scope(current_loop.into(), async move {
+//!     pyo3_asyncio::tokio::future_into_py_with_locals(
+//!         py,
+//!         locals.clone(),
+//!         // Store the current locals in task-local data
+//!         pyo3_asyncio::tokio::scope(locals.clone(), async move {
 //!             let py_sleep = Python::with_gil(|py| {
-//!                 pyo3_asyncio::into_future_with_loop(
-//!                     // Now we can get the current loop through task-local data
-//!                     pyo3_asyncio::tokio::get_current_loop(py)?,
+//!                 pyo3_asyncio::into_future_with_locals(
+//!                     // Now we can get the current locals through task-local data
+//!                     &pyo3_asyncio::tokio::get_current_locals(py)?,
 //!                     py.import("asyncio")?.call_method1("sleep", (1,))?
 //!                 )
 //!             })?;
@@ -133,18 +183,19 @@
 //! #[pyfunction]
 //! fn wrap_sleep(py: Python) -> PyResult<&PyAny> {
 //!     // get the current event loop through task-local data
-//!     // OR `asyncio.get_running_loop`
-//!     let current_loop = pyo3_asyncio::tokio::get_current_loop(py)?;
+//!     // OR `asyncio.get_running_loop` and `contextvars.copy_context`
+//!     let locals = pyo3_asyncio::tokio::get_current_locals(py)?;
 //!
-//!     pyo3_asyncio::tokio::future_into_py_with_loop(
-//!         current_loop,
-//!         // Store the current loop in task-local data
-//!         pyo3_asyncio::tokio::scope(current_loop.into(), async move {
+//!     pyo3_asyncio::tokio::future_into_py_with_locals(
+//!         py,
+//!         locals.clone(),
+//!         // Store the current locals in task-local data
+//!         pyo3_asyncio::tokio::scope(locals.clone(), async move {
 //!             let py_sleep = Python::with_gil(|py| {
-//!                 pyo3_asyncio::into_future_with_loop(
-//!                     pyo3_asyncio::tokio::get_current_loop(py)?,
+//!                 pyo3_asyncio::into_future_with_locals(
+//!                     &pyo3_asyncio::tokio::get_current_locals(py)?,
 //!                     // We can also call sleep within a Rust task since the
-//!                     // event loop is stored in task local data
+//!                     // locals are stored in task local data
 //!                     sleep(py)?
 //!                 )
 //!             })?;
@@ -165,16 +216,23 @@
 //! }
 //! ```
 //!
-//! Even though this is more correct, it's clearly not more ergonomic. That's why we introduced a new set of functions with this functionality baked in:
+//! Even though this is more correct, it's clearly not more ergonomic. That's why we introduced a
+//! set of functions with this functionality baked in:
 //!
 //! - `pyo3_asyncio::<runtime>::into_future`
-//!   > Convert a Python awaitable into a Rust future (using `pyo3_asyncio::<runtime>::get_current_loop`)
+//!   > Convert a Python awaitable into a Rust future (using
+//!   `pyo3_asyncio::<runtime>::get_current_locals`)
 //! - `pyo3_asyncio::<runtime>::future_into_py`
-//!   > Convert a Rust future into a Python awaitable (using `pyo3_asyncio::<runtime>::get_current_loop` and `pyo3_asyncio::<runtime>::scope` to set the task-local event loop for the given Rust future)
+//!   > Convert a Rust future into a Python awaitable (using
+//!   `pyo3_asyncio::<runtime>::get_current_locals` and `pyo3_asyncio::<runtime>::scope` to set the
+//!   task-local event loop for the given Rust future)
 //! - `pyo3_asyncio::<runtime>::local_future_into_py`
-//!   > Convert a `!Send` Rust future into a Python awaitable (using `pyo3_asyncio::<runtime>::get_current_loop` and `pyo3_asyncio::<runtime>::scope_local` to set the task-local event loop for the given Rust future).
+//!   > Convert a `!Send` Rust future into a Python awaitable (using
+//!   `pyo3_asyncio::<runtime>::get_current_locals` and `pyo3_asyncio::<runtime>::scope_local` to
+//!   set the task-local event loop for the given Rust future).
 //!
-//! __These are the functions that we recommend using__. With these functions, the previous example can be rewritten to be more compact:
+//! __These are the functions that we recommend using__. With these functions, the previous example
+//! can be rewritten to be more compact:
 //!
 //! ```rust
 //! use pyo3::prelude::*;
@@ -218,25 +276,17 @@
 //! }
 //! ```
 //!
-//! ## A Note for `v0.13` Users
-//!
-//! Hey guys, I realize that these are pretty major changes for `v0.14`, and I apologize in advance for having to modify the public API so much. I hope
-//! the explanation above gives some much needed context and justification for all the breaking changes.
-//!
-//! Part of the reason why it's taken so long to push out a `v0.14` release is because I wanted to make sure we got this release right. There were a lot of issues with the `v0.13` release that I hadn't anticipated, and it's thanks to your feedback and patience that we've worked through these issues to get a more correct, more flexible version out there!
-//!
-//! This new release should address most the core issues that users have reported in the `v0.13` release, so I think we can expect more stability going forward.
-//!
-//! Also, a special thanks to [@ShadowJonathan](https://github.com/ShadowJonathan) for helping with the design and review
-//! of these changes!
-//!
-//! - [@awestlake87](https://github.com/awestlake87)
+//! > A special thanks to [@ShadowJonathan](https://github.com/ShadowJonathan) for helping with the
+//! design and review of these changes!
 //!
 //! ## Rust's Event Loop
 //!
-//! Currently only the async-std and Tokio runtimes are supported by this crate.
+//! Currently only the Async-Std and Tokio runtimes are supported by this crate. If you need support
+//! for another runtime, feel free to make a request on GitHub (or attempt to add support yourself
+//! with the [`generic`] module)!
 //!
-//! > _In the future, more runtimes may be supported for Rust._
+//! > _In the future, we may implement first class support for more Rust runtimes. Contributions are
+//! welcome as well!_
 //!
 //! ## Features
 //!
@@ -249,7 +299,7 @@
 //!
 //! ```toml
 //! [dependencies.pyo3-asyncio]
-//! version = "0.13"
+//! version = "0.15"
 //! features = ["attributes"]
 //! ```
 //!
@@ -262,7 +312,7 @@
 //!
 //! ```toml
 //! [dependencies.pyo3-asyncio]
-//! version = "0.13"
+//! version = "0.15"
 //! features = ["async-std-runtime"]
 //! ```
 //!
@@ -275,7 +325,7 @@
 //!
 //! ```toml
 //! [dependencies.pyo3-asyncio]
-//! version = "0.13"
+//! version = "0.15"
 //! features = ["tokio-runtime"]
 //! ```
 //!
@@ -288,10 +338,14 @@
 //!
 //! ```toml
 //! [dependencies.pyo3-asyncio]
-//! version = "0.13"
+//! version = "0.15"
 //! features = ["testing"]
 //! ```
 
+/// Re-exported for #[test] attributes
+#[cfg(all(feature = "attributes", feature = "testing"))]
+pub use inventory;
+
 /// <span class="module-item stab portability" style="display: inline; border-radius: 3px; padding: 2px; font-size: 80%; line-height: 1.2;"><code>testing</code></span> Utilities for writing PyO3 Asyncio tests
 #[cfg(feature = "testing")]
 pub mod testing;
@@ -310,16 +364,6 @@ pub mod err;
 /// Generic implementations of PyO3 Asyncio utilities that can be used for any Rust runtime
 pub mod generic;
 
-use std::future::Future;
-
-use futures::channel::oneshot;
-use once_cell::sync::OnceCell;
-use pyo3::{exceptions::PyKeyboardInterrupt, prelude::*, types::PyTuple, PyNativeType};
-
-/// Re-exported for #[test] attributes
-#[cfg(all(feature = "attributes", feature = "testing"))]
-pub use inventory;
-
 /// Test README
 #[doc(hidden)]
 pub mod doc_test {
@@ -346,14 +390,20 @@ pub mod doc_test {
     doctest!("../README.md", readme_md);
 }
 
+use std::future::Future;
+
+use futures::channel::oneshot;
+use once_cell::sync::OnceCell;
+use pyo3::{
+    prelude::*,
+    types::{PyDict, PyTuple},
+};
+
 static ASYNCIO: OnceCell<PyObject> = OnceCell::new();
+static CONTEXTVARS: OnceCell<Option<PyObject>> = OnceCell::new();
 static ENSURE_FUTURE: OnceCell<PyObject> = OnceCell::new();
 static GET_RUNNING_LOOP: OnceCell<PyObject> = OnceCell::new();
 
-const EXPECT_INIT: &str = "PyO3 Asyncio has not been initialized";
-static CACHED_EVENT_LOOP: OnceCell<PyObject> = OnceCell::new();
-static EXECUTOR: OnceCell<PyObject> = OnceCell::new();
-
 fn ensure_future<'p>(py: Python<'p>, awaitable: &'p PyAny) -> PyResult<&'p PyAny> {
     ENSURE_FUTURE
         .get_or_try_init(|| -> PyResult<PyObject> {
@@ -367,54 +417,6 @@ fn create_future(event_loop: &PyAny) -> PyResult<&PyAny> {
     event_loop.call_method0("create_future")
 }
 
-#[allow(clippy::needless_doctest_main)]
-/// Wraps the provided function with the initialization and finalization for PyO3 Asyncio
-///
-/// This function **_MUST_** be called from the main thread.
-///
-/// # Arguments
-/// * `py` - The current PyO3 GIL guard
-/// * `f` - The function to call in between intialization and finalization
-///
-/// # Examples
-///
-/// ```
-/// use pyo3::prelude::*;
-///
-/// fn main() {
-///     // Call this function or use pyo3's "auto-initialize" feature
-///     pyo3::prepare_freethreaded_python();
-///
-///     Python::with_gil(|py| {
-///         pyo3_asyncio::with_runtime(py, || {
-///             println!("PyO3 Asyncio Initialized!");
-///             Ok(())
-///         })
-///         .map_err(|e| {
-///             e.print_and_set_sys_last_vars(py);  
-///         })
-///         .unwrap();
-///     })
-/// }
-/// ```
-#[deprecated(
-    since = "0.14.0",
-    note = "Use the pyo3_asyncio::async_std::run or pyo3_asyncio::tokio::run instead\n    (see the [migration guide](https://github.com/awestlake87/pyo3-asyncio/#migrating-from-013-to-014) for more details)"
-)]
-#[allow(deprecated)]
-pub fn with_runtime<F, R>(py: Python, f: F) -> PyResult<R>
-where
-    F: FnOnce() -> PyResult<R>,
-{
-    try_init(py)?;
-
-    let result = (f)()?;
-
-    try_close(py)?;
-
-    Ok(result)
-}
-
 fn close(event_loop: &PyAny) -> PyResult<()> {
     event_loop.call_method1(
         "run_until_complete",
@@ -434,40 +436,12 @@ fn close(event_loop: &PyAny) -> PyResult<()> {
     Ok(())
 }
 
-/// Attempt to initialize the Python and Rust event loops
-///
-/// - Must be called before any other pyo3-asyncio functions.
-/// - Calling `try_init` a second time returns `Ok(())` and does nothing.
-///   > In future versions this may return an `Err`.
-#[deprecated(
-    since = "0.14.0",
-    note = "see the [migration guide](https://github.com/awestlake87/pyo3-asyncio/#migrating-from-013-to-014) for more details"
-)]
-pub fn try_init(py: Python) -> PyResult<()> {
-    CACHED_EVENT_LOOP.get_or_try_init(|| -> PyResult<PyObject> {
-        let event_loop = asyncio_get_event_loop(py)?;
-        let executor = py
-            .import("concurrent.futures.thread")?
-            .call_method0("ThreadPoolExecutor")?;
-        event_loop.call_method1("set_default_executor", (executor,))?;
-
-        EXECUTOR.set(executor.into()).unwrap();
-        Ok(event_loop.into())
-    })?;
-
-    Ok(())
-}
-
 fn asyncio(py: Python) -> PyResult<&PyAny> {
     ASYNCIO
         .get_or_try_init(|| Ok(py.import("asyncio")?.into()))
         .map(|asyncio| asyncio.as_ref(py))
 }
 
-fn asyncio_get_event_loop(py: Python) -> PyResult<&PyAny> {
-    asyncio(py)?.call_method0("get_event_loop")
-}
-
 /// Get a reference to the Python Event Loop from Rust
 ///
 /// Equivalent to `asyncio.get_running_loop()` in Python 3.7+.
@@ -493,99 +467,76 @@ pub fn get_running_loop(py: Python) -> PyResult<&PyAny> {
         .call0()
 }
 
-/// Get a reference to the Python event loop cached by `try_init` (0.13 behaviour)
-#[deprecated(
-    since = "0.14.0",
-    note = "see the [migration guide](https://github.com/awestlake87/pyo3-asyncio/#migrating-from-013-to-014) for more details"
-)]
-pub fn get_event_loop(py: Python) -> &PyAny {
-    CACHED_EVENT_LOOP.get().expect(EXPECT_INIT).as_ref(py)
+/// Returns None only if contextvars cannot be imported (Python 3.6 fallback)
+fn contextvars(py: Python) -> Option<&PyAny> {
+    CONTEXTVARS
+        .get_or_init(|| match py.import("contextvars") {
+            Ok(contextvars) => Some(contextvars.into()),
+            Err(_) => None,
+        })
+        .as_ref()
+        .map(|contextvars| contextvars.as_ref(py))
 }
 
-/// Run the event loop forever
-///
-/// This can be called instead of `run_until_complete` to run the event loop
-/// until `stop` is called rather than driving a future to completion.
-///
-/// After this function returns, the event loop can be resumed with either `run_until_complete` or
-/// [`crate::run_forever`]
-///
-/// # Arguments
-/// * `py` - The current PyO3 GIL guard
-///
-/// # Examples
-///
-/// ```
-/// # #[cfg(feature = "async-std-runtime")]
-/// fn main() -> pyo3::PyResult<()> {
-///     use std::time::Duration;
-///     use pyo3::prelude::*;
-///     
-///     // call this or use pyo3 0.14 "auto-initialize" feature
-///     pyo3::prepare_freethreaded_python();
-///
-///     Python::with_gil(|py| {
-///         pyo3_asyncio::with_runtime(py, || {
-///             let event_loop_hdl = PyObject::from(pyo3_asyncio::get_event_loop(py));
-///             // Wait 1 second, then stop the event loop
-///             async_std::task::spawn(async move {
-///                 async_std::task::sleep(Duration::from_secs(1)).await;
-///                 Python::with_gil(|py| {
-///                     event_loop_hdl
-///                         .as_ref(py)
-///                         .call_method1(
-///                             "call_soon_threadsafe",
-///                             (event_loop_hdl
-///                                 .as_ref(py)
-///                                 .getattr("stop")
-///                                 .map_err(|e| e.print_and_set_sys_last_vars(py))
-///                                 .unwrap(),),
-///                             )
-///                             .unwrap();
-///                 })
-///             });
-///     
-///             pyo3_asyncio::run_forever(py)?;
-///
-///             Ok(())
-///         })
-///     })
-/// }
-/// # #[cfg(not(feature = "async-std-runtime"))]
-/// # fn main() {}
-#[deprecated(
-    since = "0.14.0",
-    note = "see the [migration guide](https://github.com/awestlake87/pyo3-asyncio/#migrating-from-013-to-014) for more details"
-)]
-#[allow(deprecated)]
-pub fn run_forever(py: Python) -> PyResult<()> {
-    if let Err(e) = get_event_loop(py).call_method0("run_forever") {
-        if e.is_instance::<PyKeyboardInterrupt>(py) {
-            Ok(())
-        } else {
-            Err(e)
-        }
+/// Returns Ok(None) only if contextvars cannot be imported (Python 3.6 fallback)
+fn copy_context(py: Python) -> PyResult<Option<&PyAny>> {
+    if let Some(contextvars) = contextvars(py) {
+        Ok(Some(contextvars.call_method0("copy_context")?))
     } else {
-        Ok(())
+        Ok(None)
     }
 }
 
-/// Shutdown the event loops and perform any necessary cleanup
-#[deprecated(
-    since = "0.14.0",
-    note = "see the [migration guide](https://github.com/awestlake87/pyo3-asyncio/#migrating-from-013-to-014) for more details"
-)]
-pub fn try_close(py: Python) -> PyResult<()> {
-    if let Some(exec) = EXECUTOR.get() {
-        // Shutdown the executor and wait until all threads are cleaned up
-        exec.call_method0(py, "shutdown")?;
+/// Task-local data to store for Python conversions.
+#[derive(Debug, Clone)]
+pub struct TaskLocals {
+    /// Track the event loop of the Python task
+    event_loop: PyObject,
+    /// Track the contextvars of the Python task
+    context: PyObject,
+}
+
+impl TaskLocals {
+    /// At a minimum, TaskLocals must store the event loop.
+    pub fn new(event_loop: &PyAny) -> Self {
+        Self {
+            event_loop: event_loop.into(),
+            context: event_loop.py().None(),
+        }
     }
 
-    if let Some(event_loop) = CACHED_EVENT_LOOP.get() {
-        close(event_loop.as_ref(py))?;
+    /// Construct TaskLocals with the event loop returned by `get_running_loop`
+    pub fn with_running_loop(py: Python) -> PyResult<Self> {
+        Ok(Self::new(get_running_loop(py)?))
     }
 
-    Ok(())
+    /// Manually provide the contextvars for the current task.
+    pub fn with_context(self, context: &PyAny) -> Self {
+        Self {
+            context: context.into(),
+            ..self
+        }
+    }
+
+    /// Capture the current task's contextvars
+    pub fn copy_context(self, py: Python) -> PyResult<Self> {
+        // No-op if context cannot be copied (Python 3.6 fallback)
+        if let Some(cx) = copy_context(py)? {
+            Ok(self.with_context(cx))
+        } else {
+            Ok(self)
+        }
+    }
+
+    /// Get a reference to the event loop
+    pub fn event_loop<'p>(&self, py: Python<'p>) -> &'p PyAny {
+        self.event_loop.clone().into_ref(py)
+    }
+
+    /// Get a reference to the python context
+    pub fn context<'p>(&self, py: Python<'p>) -> &'p PyAny {
+        self.context.clone().into_ref(py)
+    }
 }
 
 #[pyclass]
@@ -595,7 +546,6 @@ struct PyTaskCompleter {
 
 #[pymethods]
 impl PyTaskCompleter {
-    #[call]
     #[args(task)]
     pub fn __call__(&mut self, task: &PyAny) -> PyResult<()> {
         debug_assert!(task.call_method0("done")?.extract()?);
@@ -627,7 +577,6 @@ struct PyEnsureFuture {
 
 #[pymethods]
 impl PyEnsureFuture {
-    #[call]
     pub fn __call__(&mut self) -> PyResult<()> {
         Python::with_gil(|py| {
             let task = ensure_future(py, self.awaitable.as_ref(py))?;
@@ -639,8 +588,22 @@ impl PyEnsureFuture {
     }
 }
 
-fn call_soon_threadsafe(event_loop: &PyAny, args: impl IntoPy<Py<PyTuple>>) -> PyResult<()> {
-    event_loop.call_method1("call_soon_threadsafe", args)?;
+fn call_soon_threadsafe(
+    event_loop: &PyAny,
+    context: &PyAny,
+    args: impl IntoPy<Py<PyTuple>>,
+) -> PyResult<()> {
+    let py = event_loop.py();
+
+    let kwargs = PyDict::new(py);
+
+    // Accommodate for the Python 3.6 fallback
+    // (call_soon_threadsafe does not support the context kwarg in 3.6)
+    if !context.is_none() {
+        kwargs.set_item("context", context)?;
+    }
+
+    event_loop.call_method("call_soon_threadsafe", args, Some(kwargs))?;
     Ok(())
 }
 
@@ -652,6 +615,86 @@ fn call_soon_threadsafe(event_loop: &PyAny, args: impl IntoPy<Py<PyTuple>>) -> P
 /// simply awaits the result through the `futures::channel::oneshot::Receiver<PyResult<PyObject>>`.
 ///
 /// # Arguments
+/// * `locals` - The Python event loop and context to be used for the provided awaitable
+/// * `awaitable` - The Python `awaitable` to be converted
+///
+/// # Examples
+///
+/// ```
+/// use std::time::Duration;
+///
+/// use pyo3::prelude::*;
+///
+/// const PYTHON_CODE: &'static str = r#"
+/// import asyncio
+///
+/// async def py_sleep(duration):
+///     await asyncio.sleep(duration)
+/// "#;
+///
+/// async fn py_sleep(seconds: f32) -> PyResult<()> {
+///     let test_mod = Python::with_gil(|py| -> PyResult<PyObject> {
+///         Ok(
+///             PyModule::from_code(
+///                 py,
+///                 PYTHON_CODE,
+///                 "test_into_future/test_mod.py",
+///                 "test_mod"
+///             )?
+///             .into()
+///         )
+///     })?;
+///
+///     Python::with_gil(|py| {
+///         pyo3_asyncio::into_future_with_loop(
+///             pyo3_asyncio::get_running_loop(py)?,
+///             test_mod
+///                 .call_method1(py, "py_sleep", (seconds.into_py(py),))?
+///                 .as_ref(py),
+///         )
+///     })?
+///     .await?;
+///     Ok(())    
+/// }
+/// ```
+pub fn into_future_with_locals(
+    locals: &TaskLocals,
+    awaitable: &PyAny,
+) -> PyResult<impl Future<Output = PyResult<PyObject>> + Send> {
+    let py = awaitable.py();
+    let (tx, rx) = oneshot::channel();
+
+    call_soon_threadsafe(
+        locals.event_loop(py),
+        locals.context(py),
+        (PyEnsureFuture {
+            awaitable: awaitable.into(),
+            tx: Some(tx),
+        },),
+    )?;
+
+    Ok(async move {
+        match rx.await {
+            Ok(item) => item,
+            Err(_) => Python::with_gil(|py| {
+                Err(PyErr::from_instance(
+                    asyncio(py)?.call_method0("CancelledError")?,
+                ))
+            }),
+        }
+    })
+}
+
+/// Convert a Python `awaitable` into a Rust Future
+///
+/// This function converts the `awaitable` into a Python Task using `run_coroutine_threadsafe`. A
+/// completion handler sends the result of this Task through a
+/// `futures::channel::oneshot::Sender<PyResult<PyObject>>` and the future returned by this function
+/// simply awaits the result through the `futures::channel::oneshot::Receiver<PyResult<PyObject>>`.
+///
+/// __This function will be removed in `v0.16`__
+///
+/// # Arguments
 /// * `event_loop` - The Python event loop that the awaitable should be attached to
 /// * `awaitable` - The Python `awaitable` to be converted
 ///
@@ -694,88 +737,18 @@ fn call_soon_threadsafe(event_loop: &PyAny, args: impl IntoPy<Py<PyTuple>>) -> P
 ///     Ok(())    
 /// }
 /// ```
+#[deprecated(
+    since = "0.15.0",
+    note = "Use pyo3_asyncio::into_future_with_locals instead"
+)]
 pub fn into_future_with_loop(
     event_loop: &PyAny,
     awaitable: &PyAny,
 ) -> PyResult<impl Future<Output = PyResult<PyObject>> + Send> {
-    let (tx, rx) = oneshot::channel();
-
-    call_soon_threadsafe(
-        event_loop,
-        (PyEnsureFuture {
-            awaitable: awaitable.into(),
-            tx: Some(tx),
-        },),
-    )?;
-
-    Ok(async move {
-        match rx.await {
-            Ok(item) => item,
-            Err(_) => Python::with_gil(|py| {
-                Err(PyErr::from_instance(
-                    asyncio(py)?.call_method0("CancelledError")?,
-                ))
-            }),
-        }
-    })
-}
-
-/// Convert a Python `awaitable` into a Rust Future
-///
-/// This function converts the `awaitable` into a Python Task using `run_coroutine_threadsafe`. A
-/// completion handler sends the result of this Task through a
-/// `futures::channel::oneshot::Sender<PyResult<PyObject>>` and the future returned by this function
-/// simply awaits the result through the `futures::channel::oneshot::Receiver<PyResult<PyObject>>`.
-///
-/// # Arguments
-/// * `awaitable` - The Python `awaitable` to be converted
-///
-/// # Examples
-///
-/// ```
-/// use std::time::Duration;
-///
-/// use pyo3::prelude::*;
-///
-/// const PYTHON_CODE: &'static str = r#"
-/// import asyncio
-///
-/// async def py_sleep(duration):
-///     await asyncio.sleep(duration)
-/// "#;
-///
-/// async fn py_sleep(seconds: f32) -> PyResult<()> {
-///     let test_mod = Python::with_gil(|py| -> PyResult<PyObject> {
-///         Ok(
-///             PyModule::from_code(
-///                 py,
-///                 PYTHON_CODE,
-///                 "test_into_future/test_mod.py",
-///                 "test_mod"
-///             )?
-///             .into()
-///         )
-///     })?;
-///
-///     Python::with_gil(|py| {
-///         // Only works with cached event loop
-///         pyo3_asyncio::into_future(
-///             test_mod
-///                 .call_method1(py, "py_sleep", (seconds.into_py(py),))?
-///                 .as_ref(py),
-///         )
-///     })?
-///     .await?;
-///     Ok(())    
-/// }
-/// ```
-#[deprecated(
-    since = "0.14.0",
-    note = "Use pyo3_asyncio::async_std::into_future or pyo3_asyncio::tokio::into_future instead\n    (see the [migration guide](https://github.com/awestlake87/pyo3-asyncio/#migrating-from-013-to-014) for more details)"
-)]
-#[allow(deprecated)]
-pub fn into_future(awaitable: &PyAny) -> PyResult<impl Future<Output = PyResult<PyObject>> + Send> {
-    into_future_with_loop(get_event_loop(awaitable.py()), awaitable)
+    into_future_with_locals(
+        &TaskLocals::new(event_loop).copy_context(event_loop.py())?,
+        awaitable,
+    )
 }
 
 fn dump_err(py: Python<'_>) -> impl FnOnce(PyErr) + '_ {
