@@ -10,7 +10,10 @@ use once_cell::{
 };
 use pyo3::prelude::*;
 
-use crate::generic::{self, Runtime as GenericRuntime, SpawnLocalExt};
+use crate::{
+    generic::{self, ContextExt, LocalContextExt, Runtime as GenericRuntime, SpawnLocalExt},
+    TaskLocals,
+};
 
 /// <span class="module-item stab portability" style="display: inline; border-radius: 3px; padding: 2px; font-size: 80%; line-height: 1.2;"><code>attributes</code></span>
 /// re-exports for macros
@@ -44,30 +47,13 @@ impl generic::JoinError for task::JoinError {
 struct TokioRuntime;
 
 tokio::task_local! {
-    static EVENT_LOOP: UnsyncOnceCell<PyObject>;
+    static TASK_LOCALS: UnsyncOnceCell<TaskLocals>;
 }
 
 impl GenericRuntime for TokioRuntime {
     type JoinError = task::JoinError;
     type JoinHandle = task::JoinHandle<()>;
 
-    fn scope<F, R>(event_loop: PyObject, fut: F) -> Pin<Box<dyn Future<Output = R> + Send>>
-    where
-        F: Future<Output = R> + Send + 'static,
-    {
-        let cell = UnsyncOnceCell::new();
-        cell.set(event_loop).unwrap();
-
-        Box::pin(EVENT_LOOP.scope(cell, fut))
-    }
-
-    fn get_task_event_loop(py: Python) -> Option<&PyAny> {
-        match EVENT_LOOP.try_with(|c| c.get().map(|event_loop| event_loop.clone().into_ref(py))) {
-            Ok(event_loop) => event_loop,
-            Err(_) => None,
-        }
-    }
-
     fn spawn<F>(fut: F) -> Self::JoinHandle
     where
         F: Future<Output = ()> + Send + 'static,
@@ -78,17 +64,26 @@ impl GenericRuntime for TokioRuntime {
     }
 }
 
-impl SpawnLocalExt for TokioRuntime {
-    fn scope_local<F, R>(event_loop: PyObject, fut: F) -> Pin<Box<dyn Future<Output = R>>>
+impl ContextExt for TokioRuntime {
+    fn scope<F, R>(locals: TaskLocals, fut: F) -> Pin<Box<dyn Future<Output = R> + Send>>
     where
-        F: Future<Output = R> + 'static,
+        F: Future<Output = R> + Send + 'static,
     {
         let cell = UnsyncOnceCell::new();
-        cell.set(event_loop).unwrap();
+        cell.set(locals).unwrap();
 
-        Box::pin(EVENT_LOOP.scope(cell, fut))
+        Box::pin(TASK_LOCALS.scope(cell, fut))
     }
 
+    fn get_task_locals() -> Option<TaskLocals> {
+        match TASK_LOCALS.try_with(|c| c.get().map(|locals| locals.clone())) {
+            Ok(locals) => locals,
+            Err(_) => None,
+        }
+    }
+}
+
+impl SpawnLocalExt for TokioRuntime {
     fn spawn_local<F>(fut: F) -> Self::JoinHandle
     where
         F: Future<Output = ()> + 'static,
@@ -97,20 +92,32 @@ impl SpawnLocalExt for TokioRuntime {
     }
 }
 
+impl LocalContextExt for TokioRuntime {
+    fn scope_local<F, R>(locals: TaskLocals, fut: F) -> Pin<Box<dyn Future<Output = R>>>
+    where
+        F: Future<Output = R> + 'static,
+    {
+        let cell = UnsyncOnceCell::new();
+        cell.set(locals).unwrap();
+
+        Box::pin(TASK_LOCALS.scope(cell, fut))
+    }
+}
+
 /// Set the task local event loop for the given future
-pub async fn scope<F, R>(event_loop: PyObject, fut: F) -> R
+pub async fn scope<F, R>(locals: TaskLocals, fut: F) -> R
 where
     F: Future<Output = R> + Send + 'static,
 {
-    TokioRuntime::scope(event_loop, fut).await
+    TokioRuntime::scope(locals, fut).await
 }
 
 /// Set the task local event loop for the given !Send future
-pub async fn scope_local<F, R>(event_loop: PyObject, fut: F) -> R
+pub async fn scope_local<F, R>(locals: TaskLocals, fut: F) -> R
 where
     F: Future<Output = R> + 'static,
 {
-    TokioRuntime::scope_local(event_loop, fut).await
+    TokioRuntime::scope_local(locals, fut).await
 }
 
 /// Get the current event loop from either Python or Rust async task local context
@@ -122,6 +129,12 @@ pub fn get_current_loop(py: Python) -> PyResult<&PyAny> {
     generic::get_current_loop::<TokioRuntime>(py)
 }
 
+/// Either copy the task locals from the current task OR get the current running loop and
+/// contextvars from Python.
+pub fn get_current_locals(py: Python) -> PyResult<TaskLocals> {
+    generic::get_current_locals::<TokioRuntime>(py)
+}
+
 /// Initialize the Tokio runtime with a custom build
 pub fn init(builder: Builder) {
     *TOKIO_BUILDER.lock().unwrap() = builder
@@ -148,8 +161,7 @@ fn multi_thread() -> Builder {
 ///
 /// The event loop runs until the given future is complete.
 ///
-/// After this function returns, the event loop can be resumed with either [`run_until_complete`] or
-/// [`crate::run_forever`]
+/// After this function returns, the event loop can be resumed with [`run_until_complete`]
 ///
 /// # Arguments
 /// * `event_loop` - The Python event loop that should run the future
@@ -163,26 +175,21 @@ fn multi_thread() -> Builder {
 /// # use pyo3::prelude::*;
 /// #
 /// # pyo3::prepare_freethreaded_python();
-/// # Python::with_gil(|py| {
-/// # pyo3_asyncio::with_runtime(py, || {
+/// # Python::with_gil(|py| -> PyResult<()> {
 /// # let event_loop = py.import("asyncio")?.call_method0("new_event_loop")?;
 /// pyo3_asyncio::tokio::run_until_complete(event_loop, async move {
 ///     tokio::time::sleep(Duration::from_secs(1)).await;
 ///     Ok(())
 /// })?;
 /// # Ok(())
-/// # })
-/// # .map_err(|e| {
-/// #    e.print_and_set_sys_last_vars(py);  
-/// # })
-/// # .unwrap();
-/// # });
+/// # }).unwrap();
 /// ```
-pub fn run_until_complete<F>(event_loop: &PyAny, fut: F) -> PyResult<()>
+pub fn run_until_complete<F, T>(event_loop: &PyAny, fut: F) -> PyResult<T>
 where
-    F: Future<Output = PyResult<()>> + Send + 'static,
+    F: Future<Output = PyResult<T>> + Send + 'static,
+    T: Send + Sync + 'static,
 {
-    generic::run_until_complete::<TokioRuntime, _>(event_loop, fut)
+    generic::run_until_complete::<TokioRuntime, _, T>(event_loop, fut)
 }
 
 /// Run the event loop until the given Future completes
@@ -211,49 +218,17 @@ where
 ///     })
 /// }
 /// ```
-pub fn run<F>(py: Python, fut: F) -> PyResult<()>
+pub fn run<F, T>(py: Python, fut: F) -> PyResult<T>
 where
-    F: Future<Output = PyResult<()>> + Send + 'static,
+    F: Future<Output = PyResult<T>> + Send + 'static,
+    T: Send + Sync + 'static,
 {
-    generic::run::<TokioRuntime, F>(py, fut)
+    generic::run::<TokioRuntime, F, T>(py, fut)
 }
 
 /// Convert a Rust Future into a Python awaitable
 ///
-/// # Arguments
-/// * `py` - The current PyO3 GIL guard
-/// * `fut` - The Rust future to be converted
-///
-/// # Examples
-///
-/// ```
-/// use std::time::Duration;
-///
-/// use pyo3::prelude::*;
-///
-/// /// Awaitable sleep function
-/// #[pyfunction]
-/// fn sleep_for(py: Python, secs: &PyAny) -> PyResult<PyObject> {
-///     let secs = secs.extract()?;
-///     pyo3_asyncio::tokio::into_coroutine(py, async move {
-///         tokio::time::sleep(Duration::from_secs(secs)).await;
-///         Python::with_gil(|py| Ok(py.None()))
-///     })
-/// }
-/// ```
-#[deprecated(
-    since = "0.14.0",
-    note = "Use pyo3_asyncio::tokio::future_into_py instead\n    (see the [migration guide](https://github.com/awestlake87/pyo3-asyncio/#migrating-from-013-to-014) for more details)"
-)]
-#[allow(deprecated)]
-pub fn into_coroutine<F>(py: Python, fut: F) -> PyResult<PyObject>
-where
-    F: Future<Output = PyResult<PyObject>> + Send + 'static,
-{
-    generic::into_coroutine::<TokioRuntime, _>(py, fut)
-}
-
-/// Convert a Rust Future into a Python awaitable
+/// __This function will be removed in `v0.16`__
 ///
 /// # Arguments
 /// * `event_loop` - The Python event loop that the awaitable should be attached to
@@ -279,6 +254,11 @@ where
 ///     )
 /// }
 /// ```
+#[deprecated(
+    since = "0.15.0",
+    note = "Use pyo3_asyncio::tokio::future_into_py_with_locals instead"
+)]
+#[allow(deprecated)]
 pub fn future_into_py_with_loop<F>(event_loop: &PyAny, fut: F) -> PyResult<&PyAny>
 where
     F: Future<Output = PyResult<PyObject>> + Send + 'static,
@@ -288,12 +268,62 @@ where
 
 /// Convert a Rust Future into a Python awaitable
 ///
-/// Unlike [`future_into_py_with_loop`], this function will stop the Rust future from running when
-/// the `asyncio.Future` is cancelled from Python.
+/// If the `asyncio.Future` returned by this conversion is cancelled via `asyncio.Future.cancel`,
+/// the Rust future will be cancelled as well (new behaviour in `v0.15`).
 ///
-/// __This function will be deprecated in favor of [`future_into_py_with_loop`] in `v0.15` because
-/// it will become the default behaviour. In `v0.15`, any calls to this function can be seamlessly
-/// replaced with [`future_into_py_with_loop`].__
+/// Python `contextvars` are preserved when calling async Python functions within the Rust future
+/// via [`into_future`] (new behaviour in `v0.15`).
+///
+/// > Although `contextvars` are preserved for async Python functions, synchronous functions will
+/// unfortunately fail to resolve them when called within the Rust future. This is because the
+/// function is being called from a Rust thread, not inside an actual Python coroutine context.
+/// >
+/// > As a workaround, you can get the `contextvars` from the current task locals using
+/// [`get_current_locals`] and [`TaskLocals::context`](`crate::TaskLocals::context`), then wrap your
+/// synchronous function in a call to `contextvars.Context.run`. This will set the context, call the
+/// synchronous function, and restore the previous context when it returns or raises an exception.
+///
+/// # Arguments
+/// * `py` - PyO3 GIL guard
+/// * `locals` - The task locals for the given future
+/// * `fut` - The Rust future to be converted
+///
+/// # Examples
+///
+/// ```
+/// use std::time::Duration;
+///
+/// use pyo3::prelude::*;
+///
+/// /// Awaitable sleep function
+/// #[pyfunction]
+/// fn sleep_for<'p>(py: Python<'p>, secs: &'p PyAny) -> PyResult<&'p PyAny> {
+///     let secs = secs.extract()?;
+///     pyo3_asyncio::tokio::future_into_py_with_locals(
+///         py,
+///         pyo3_asyncio::tokio::get_current_locals(py)?,
+///         async move {
+///             tokio::time::sleep(Duration::from_secs(secs)).await;
+///             Python::with_gil(|py| Ok(py.None()))
+///         }
+///     )
+/// }
+/// ```
+pub fn future_into_py_with_locals<F, T>(py: Python, locals: TaskLocals, fut: F) -> PyResult<&PyAny>
+where
+    F: Future<Output = PyResult<T>> + Send + 'static,
+    T: IntoPy<PyObject>,
+{
+    generic::future_into_py_with_locals::<TokioRuntime, F, T>(py, locals, fut)
+}
+
+/// Convert a Rust Future into a Python awaitable
+///
+/// __This function was deprecated in favor of [`future_into_py_with_locals`] in `v0.15` because
+/// it became the default behaviour. In `v0.15`, any calls to this function should be
+/// replaced with [`future_into_py_with_locals`].__
+///
+/// __This function will be removed in `v0.16`__
 ///
 /// # Arguments
 /// * `event_loop` - The Python event loop that the awaitable should be attached to
@@ -314,11 +344,16 @@ where
 ///         pyo3_asyncio::tokio::get_current_loop(py)?,
 ///         async move {
 ///             tokio::time::sleep(Duration::from_secs(secs)).await;
-///             Python::with_gil(|py| Ok(py.None()))
+///             Ok(Python::with_gil(|py| py.None()))
 ///         }
 ///     )
 /// }
 /// ```
+#[deprecated(
+    since = "0.15.0",
+    note = "Use pyo3_asyncio::tokio::future_into_py_with_locals instead"
+)]
+#[allow(deprecated)]
 pub fn cancellable_future_into_py_with_loop<F>(event_loop: &PyAny, fut: F) -> PyResult<&PyAny>
 where
     F: Future<Output = PyResult<PyObject>> + Send + 'static,
@@ -328,6 +363,21 @@ where
 
 /// Convert a Rust Future into a Python awaitable
 ///
+/// If the `asyncio.Future` returned by this conversion is cancelled via `asyncio.Future.cancel`,
+/// the Rust future will be cancelled as well (new behaviour in `v0.15`).
+///
+/// Python `contextvars` are preserved when calling async Python functions within the Rust future
+/// via [`into_future`] (new behaviour in `v0.15`).
+///
+/// > Although `contextvars` are preserved for async Python functions, synchronous functions will
+/// unfortunately fail to resolve them when called within the Rust future. This is because the
+/// function is being called from a Rust thread, not inside an actual Python coroutine context.
+/// >
+/// > As a workaround, you can get the `contextvars` from the current task locals using
+/// [`get_current_locals`] and [`TaskLocals::context`](`crate::TaskLocals::context`), then wrap your
+/// synchronous function in a call to `contextvars.Context.run`. This will set the context, call the
+/// synchronous function, and restore the previous context when it returns or raises an exception.
+///
 /// # Arguments
 /// * `py` - The current PyO3 GIL guard
 /// * `fut` - The Rust future to be converted
@@ -345,26 +395,26 @@ where
 ///     let secs = secs.extract()?;
 ///     pyo3_asyncio::tokio::future_into_py(py, async move {
 ///         tokio::time::sleep(Duration::from_secs(secs)).await;
-///         Python::with_gil(|py| Ok(py.None()))
+///         Ok(())
 ///     })
 /// }
 /// ```
-pub fn future_into_py<F>(py: Python, fut: F) -> PyResult<&PyAny>
+pub fn future_into_py<F, T>(py: Python, fut: F) -> PyResult<&PyAny>
 where
-    F: Future<Output = PyResult<PyObject>> + Send + 'static,
+    F: Future<Output = PyResult<T>> + Send + 'static,
+    T: IntoPy<PyObject>,
 {
-    generic::future_into_py::<TokioRuntime, _>(py, fut)
+    generic::future_into_py::<TokioRuntime, _, T>(py, fut)
 }
 
 /// Convert a Rust Future into a Python awaitable
 ///
-/// Unlike [`future_into_py`], this function will stop the Rust future from running when
-/// the `asyncio.Future` is cancelled from Python.
-///
-/// __This function will be deprecated in favor of [`future_into_py`] in `v0.15` because
-/// it will become the default behaviour. In `v0.15`, any calls to this function can be seamlessly
+/// __This function was deprecated in favor of [`future_into_py`] in `v0.15` because
+/// it became the default behaviour. In `v0.15`, any calls to this function can be seamlessly
 /// replaced with [`future_into_py`].__
 ///
+/// __This function will be removed in `v0.16`__
+///
 /// # Arguments
 /// * `py` - The current PyO3 GIL guard
 /// * `fut` - The Rust future to be converted
@@ -386,6 +436,11 @@ where
 ///     })
 /// }
 /// ```
+#[deprecated(
+    since = "0.15.0",
+    note = "Use pyo3_asyncio::tokio::future_into_py instead"
+)]
+#[allow(deprecated)]
 pub fn cancellable_future_into_py<F>(py: Python, fut: F) -> PyResult<&PyAny>
 where
     F: Future<Output = PyResult<PyObject>> + Send + 'static,
@@ -395,6 +450,8 @@ where
 
 /// Convert a `!Send` Rust Future into a Python awaitable
 ///
+/// __This function will be removed in `v0.16`__
+///
 /// # Arguments
 /// * `event_loop` - The Python event loop that the awaitable should be attached to
 /// * `fut` - The Rust future to be converted
@@ -416,7 +473,7 @@ where
 ///         pyo3_asyncio::tokio::get_current_loop(py)?,
 ///         async move {
 ///             tokio::time::sleep(Duration::from_secs(*secs)).await;
-///             Python::with_gil(|py| Ok(py.None()))
+///             Ok(Python::with_gil(|py| py.None()))
 ///         }
 ///     )
 /// }
@@ -424,8 +481,8 @@ where
 /// # #[cfg(all(feature = "tokio-runtime", feature = "attributes"))]
 /// #[pyo3_asyncio::tokio::main]
 /// async fn main() -> PyResult<()> {
-///     let event_loop = Python::with_gil(|py| -> PyResult<PyObject> {
-///         Ok(pyo3_asyncio::tokio::get_current_loop(py)?.into())
+///     let locals = Python::with_gil(|py| -> PyResult<_> {
+///         pyo3_asyncio::tokio::get_current_locals(py)
 ///     })?;
 ///
 ///     // the main coroutine is running in a Send context, so we cannot use LocalSet here. Instead
@@ -435,7 +492,7 @@ where
 ///         // pyo3_asyncio::tokio::local_future_into_py will panic.
 ///         tokio::task::LocalSet::new().block_on(
 ///             pyo3_asyncio::tokio::get_runtime(),  
-///             pyo3_asyncio::tokio::scope_local(event_loop, async {
+///             pyo3_asyncio::tokio::scope_local(locals, async {
 ///                 Python::with_gil(|py| {
 ///                     let py_future = sleep_for(py, 1)?;
 ///                     pyo3_asyncio::tokio::into_future(py_future)
@@ -450,6 +507,11 @@ where
 /// # #[cfg(not(all(feature = "tokio-runtime", feature = "attributes")))]
 /// # fn main() {}
 /// ```
+#[deprecated(
+    since = "0.15.0",
+    note = "Use pyo3_asyncio::tokio::local_future_into_py_with_locals instead"
+)]
+#[allow(deprecated)]
 pub fn local_future_into_py_with_loop<'p, F>(event_loop: &'p PyAny, fut: F) -> PyResult<&PyAny>
 where
     F: Future<Output = PyResult<PyObject>> + 'static,
@@ -459,12 +521,97 @@ where
 
 /// Convert a `!Send` Rust Future into a Python awaitable
 ///
-/// Unlike [`local_future_into_py_with_loop`], this function will stop the Rust future from running when
-/// the `asyncio.Future` is cancelled from Python.
+/// If the `asyncio.Future` returned by this conversion is cancelled via `asyncio.Future.cancel`,
+/// the Rust future will be cancelled as well (new behaviour in `v0.15`).
 ///
-/// __This function will be deprecated in favor of [`local_future_into_py_with_loop`] in `v0.15` because
-/// it will become the default behaviour. In `v0.15`, any calls to this function can be seamlessly
-/// replaced with [`local_future_into_py_with_loop`].__
+/// Python `contextvars` are preserved when calling async Python functions within the Rust future
+/// via [`into_future`] (new behaviour in `v0.15`).
+///
+/// > Although `contextvars` are preserved for async Python functions, synchronous functions will
+/// unfortunately fail to resolve them when called within the Rust future. This is because the
+/// function is being called from a Rust thread, not inside an actual Python coroutine context.
+/// >
+/// > As a workaround, you can get the `contextvars` from the current task locals using
+/// [`get_current_locals`] and [`TaskLocals::context`](`crate::TaskLocals::context`), then wrap your
+/// synchronous function in a call to `contextvars.Context.run`. This will set the context, call the
+/// synchronous function, and restore the previous context when it returns or raises an exception.
+///
+/// # Arguments
+/// * `py` - PyO3 GIL guard
+/// * `locals` - The task locals for the given future
+/// * `fut` - The Rust future to be converted
+///
+/// # Examples
+///
+/// ```
+/// use std::{rc::Rc, time::Duration};
+///
+/// use pyo3::prelude::*;
+///
+/// /// Awaitable non-send sleep function
+/// #[pyfunction]
+/// fn sleep_for(py: Python, secs: u64) -> PyResult<&PyAny> {
+///     // Rc is non-send so it cannot be passed into pyo3_asyncio::tokio::future_into_py
+///     let secs = Rc::new(secs);
+///
+///     pyo3_asyncio::tokio::local_future_into_py_with_locals(
+///         py,
+///         pyo3_asyncio::tokio::get_current_locals(py)?,
+///         async move {
+///             tokio::time::sleep(Duration::from_secs(*secs)).await;
+///             Python::with_gil(|py| Ok(py.None()))
+///         }
+///     )
+/// }
+///
+/// # #[cfg(all(feature = "tokio-runtime", feature = "attributes"))]
+/// #[pyo3_asyncio::tokio::main]
+/// async fn main() -> PyResult<()> {
+///     let locals = Python::with_gil(|py| -> PyResult<_> {
+///         pyo3_asyncio::tokio::get_current_locals(py)
+///     })?;
+///
+///     // the main coroutine is running in a Send context, so we cannot use LocalSet here. Instead
+///     // we use spawn_blocking in order to use LocalSet::block_on
+///     tokio::task::spawn_blocking(move || {
+///         // LocalSet allows us to work with !Send futures within tokio. Without it, any calls to
+///         // pyo3_asyncio::tokio::local_future_into_py will panic.
+///         tokio::task::LocalSet::new().block_on(
+///             pyo3_asyncio::tokio::get_runtime(),  
+///             pyo3_asyncio::tokio::scope_local(locals, async {
+///                 Python::with_gil(|py| {
+///                     let py_future = sleep_for(py, 1)?;
+///                     pyo3_asyncio::tokio::into_future(py_future)
+///                 })?
+///                 .await?;
+///
+///                 Ok(())
+///             })
+///         )
+///     }).await.unwrap()
+/// }
+/// # #[cfg(not(all(feature = "tokio-runtime", feature = "attributes")))]
+/// # fn main() {}
+/// ```
+pub fn local_future_into_py_with_locals<F, T>(
+    py: Python,
+    locals: TaskLocals,
+    fut: F,
+) -> PyResult<&PyAny>
+where
+    F: Future<Output = PyResult<T>> + 'static,
+    T: IntoPy<PyObject>,
+{
+    generic::local_future_into_py_with_locals::<TokioRuntime, _, T>(py, locals, fut)
+}
+
+/// Convert a `!Send` Rust Future into a Python awaitable
+///
+/// __This function was deprecated in favor of [`local_future_into_py_with_locals`] in `v0.15` because
+/// it became the default behaviour. In `v0.15`, any calls to this function should be
+/// replaced with [`local_future_into_py_with_locals`].__
+///
+/// __This function will be removed in `v0.16`__
 ///
 /// # Arguments
 /// * `event_loop` - The Python event loop that the awaitable should be attached to
@@ -495,8 +642,8 @@ where
 /// # #[cfg(all(feature = "tokio-runtime", feature = "attributes"))]
 /// #[pyo3_asyncio::tokio::main]
 /// async fn main() -> PyResult<()> {
-///     let event_loop = Python::with_gil(|py| -> PyResult<PyObject> {
-///         Ok(pyo3_asyncio::tokio::get_current_loop(py)?.into())
+///     let locals = Python::with_gil(|py| -> PyResult<_> {
+///         pyo3_asyncio::tokio::get_current_locals(py)
 ///     })?;
 ///
 ///     // the main coroutine is running in a Send context, so we cannot use LocalSet here. Instead
@@ -506,7 +653,7 @@ where
 ///         // pyo3_asyncio::tokio::local_future_into_py will panic.
 ///         tokio::task::LocalSet::new().block_on(
 ///             pyo3_asyncio::tokio::get_runtime(),  
-///             pyo3_asyncio::tokio::scope_local(event_loop, async {
+///             pyo3_asyncio::tokio::scope_local(locals, async {
 ///                 Python::with_gil(|py| {
 ///                     let py_future = sleep_for(py, 1)?;
 ///                     pyo3_asyncio::tokio::into_future(py_future)
@@ -521,6 +668,11 @@ where
 /// # #[cfg(not(all(feature = "tokio-runtime", feature = "attributes")))]
 /// # fn main() {}
 /// ```
+#[deprecated(
+    since = "0.15.0",
+    note = "Use pyo3_asyncio::tokio::local_future_into_py_with_locals instead"
+)]
+#[allow(deprecated)]
 pub fn local_cancellable_future_into_py_with_loop<'p, F>(
     event_loop: &'p PyAny,
     fut: F,
@@ -533,6 +685,21 @@ where
 
 /// Convert a `!Send` Rust Future into a Python awaitable
 ///
+/// If the `asyncio.Future` returned by this conversion is cancelled via `asyncio.Future.cancel`,
+/// the Rust future will be cancelled as well (new behaviour in `v0.15`).
+///
+/// Python `contextvars` are preserved when calling async Python functions within the Rust future
+/// via [`into_future`] (new behaviour in `v0.15`).
+///
+/// > Although `contextvars` are preserved for async Python functions, synchronous functions will
+/// unfortunately fail to resolve them when called within the Rust future. This is because the
+/// function is being called from a Rust thread, not inside an actual Python coroutine context.
+/// >
+/// > As a workaround, you can get the `contextvars` from the current task locals using
+/// [`get_current_locals`] and [`TaskLocals::context`](`crate::TaskLocals::context`), then wrap your
+/// synchronous function in a call to `contextvars.Context.run`. This will set the context, call the
+/// synchronous function, and restore the previous context when it returns or raises an exception.
+///
 /// # Arguments
 /// * `py` - The current PyO3 GIL guard
 /// * `fut` - The Rust future to be converted
@@ -551,15 +718,15 @@ where
 ///     let secs = Rc::new(secs);
 ///     pyo3_asyncio::tokio::local_future_into_py(py, async move {
 ///         tokio::time::sleep(Duration::from_secs(*secs)).await;
-///         Python::with_gil(|py| Ok(py.None()))
+///         Ok(())
 ///     })
 /// }
 ///
 /// # #[cfg(all(feature = "tokio-runtime", feature = "attributes"))]
 /// #[pyo3_asyncio::tokio::main]
 /// async fn main() -> PyResult<()> {
-///     let event_loop = Python::with_gil(|py| {
-///         PyObject::from(pyo3_asyncio::tokio::get_current_loop(py).unwrap())
+///     let locals = Python::with_gil(|py| {
+///         pyo3_asyncio::tokio::get_current_locals(py).unwrap()
 ///     });
 ///
 ///     // the main coroutine is running in a Send context, so we cannot use LocalSet here. Instead
@@ -569,7 +736,7 @@ where
 ///         // pyo3_asyncio::tokio::local_future_into_py will panic.
 ///         tokio::task::LocalSet::new().block_on(
 ///             pyo3_asyncio::tokio::get_runtime(),  
-///             pyo3_asyncio::tokio::scope_local(event_loop, async {
+///             pyo3_asyncio::tokio::scope_local(locals, async {
 ///                 Python::with_gil(|py| {
 ///                     let py_future = sleep_for(py, 1)?;
 ///                     pyo3_asyncio::tokio::into_future(py_future)
@@ -584,22 +751,22 @@ where
 /// # #[cfg(not(all(feature = "tokio-runtime", feature = "attributes")))]
 /// # fn main() {}
 /// ```
-pub fn local_future_into_py<F>(py: Python, fut: F) -> PyResult<&PyAny>
+pub fn local_future_into_py<F, T>(py: Python, fut: F) -> PyResult<&PyAny>
 where
-    F: Future<Output = PyResult<PyObject>> + 'static,
+    F: Future<Output = PyResult<T>> + 'static,
+    T: IntoPy<PyObject>,
 {
-    generic::local_future_into_py::<TokioRuntime, _>(py, fut)
+    generic::local_future_into_py::<TokioRuntime, _, T>(py, fut)
 }
 
 /// Convert a `!Send` Rust Future into a Python awaitable
 ///
-/// Unlike [`local_future_into_py`], this function will stop the Rust future from running when
-/// the `asyncio.Future` is cancelled from Python.
-///
-/// __This function will be deprecated in favor of [`local_future_into_py`] in `v0.15` because
-/// it will become the default behaviour. In `v0.15`, any calls to this function can be seamlessly
+/// __This function was deprecated in favor of [`local_future_into_py`] in `v0.15` because
+/// it became the default behaviour. In `v0.15`, any calls to this function can be seamlessly
 /// replaced with [`local_future_into_py`].__
 ///
+/// __This function will be removed in `v0.16`__
+///
 /// # Arguments
 /// * `py` - The current PyO3 GIL guard
 /// * `fut` - The Rust future to be converted
@@ -625,8 +792,8 @@ where
 /// # #[cfg(all(feature = "tokio-runtime", feature = "attributes"))]
 /// #[pyo3_asyncio::tokio::main]
 /// async fn main() -> PyResult<()> {
-///     let event_loop = Python::with_gil(|py| {
-///         PyObject::from(pyo3_asyncio::tokio::get_current_loop(py).unwrap())
+///     let locals = Python::with_gil(|py| {
+///         pyo3_asyncio::tokio::get_current_locals(py).unwrap()
 ///     });
 ///
 ///     // the main coroutine is running in a Send context, so we cannot use LocalSet here. Instead
@@ -636,7 +803,7 @@ where
 ///         // pyo3_asyncio::tokio::local_future_into_py will panic.
 ///         tokio::task::LocalSet::new().block_on(
 ///             pyo3_asyncio::tokio::get_runtime(),  
-///             pyo3_asyncio::tokio::scope_local(event_loop, async {
+///             pyo3_asyncio::tokio::scope_local(locals, async {
 ///                 Python::with_gil(|py| {
 ///                     let py_future = sleep_for(py, 1)?;
 ///                     pyo3_asyncio::tokio::into_future(py_future)
@@ -651,6 +818,11 @@ where
 /// # #[cfg(not(all(feature = "tokio-runtime", feature = "attributes")))]
 /// # fn main() {}
 /// ```
+#[deprecated(
+    since = "0.15.0",
+    note = "Use pyo3_asyncio::tokio::local_future_into_py instead"
+)]
+#[allow(deprecated)]
 pub fn local_cancellable_future_into_py<F>(py: Python, fut: F) -> PyResult<&PyAny>
 where
     F: Future<Output = PyResult<PyObject>> + 'static,
